/*
** AHXPLAY v0.23 - 9th of March 2018 - https://16-bits.org
** =======================================================
**               - NOT BIG ENDIAN SAFE! -
**
** Work in progress, probably is buggy...
**
** C port of AHX.cpp (Bartman of Abyss) by Olav "8bitbubsy" Sørensen
** Includes additional fixes from hively tracker (current github code)
**
** Do NOT use a lower audio rate than 32kHz! This will cause problems
** with low sample periods (high rates) and BLEP.
**
** The BLEP (Band-Limited Step) and high-pass filter were coded by aciddose.
** This makes the replayer sound much similar to a real Amiga 1200.
**
** You need to link winmm.lib for this to compile (-lwinmm).
** Alternatively, you can wrap around your own audio callback for using
** ahxplay on other platforms (read comment near bottom).
**
**
** ahxplay.h:
**
** #include <stdint.h>
**
** int8_t ahxplay_Init(uint32_t outputFreq);
** int8_t ahxplay_LoadSong(void *Buffer, int32_t Len);
** int8_t ahxplay_Play(int32_t subSongNr);
** void ahxplay_Close(void);
** char *ahxplay_GetSongName(void);
*/


/* == USER ADJUSTABLE SETTINGS == */
#define STEREO_SEP (19)    /* --> Stereo separation in percent - 0 = mono, 100 = hard pan (like Amiga) */
#define NORM_FACTOR (4.0f) /* --> Slightly increase this value if the song is too loud. Decrease if too quiet... */
#define USE_HIGHPASS       /* --> 5.2Hz high-pass filter present in all Amigas - comment out for a speed-up */
#define USE_BLEP           /* --> Reduces some unwanted aliasing (closer to real Amiga) - comment out to disable */
#define AUDIO_BUF_LEN (4096)


#ifndef _USE_MATH_DEFINES
#define _USE_MATH_DEFINES
#endif

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#ifndef true
#define true 1
#define false 0
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <windows.h>
#include <mmsystem.h>
#include <math.h> // tanf()

/* don't modify these defines! */
#define AMIGA_VOICES 4
#define NUM_WAVEFORMS (6 + 6 + 32 + 1)
#define FILTER_ITERATIONS 31
#define BLEP_ZC 8
#define BLEP_OS 5
#define BLEP_SP 5
#define BLEP_NS (BLEP_ZC * BLEP_OS / BLEP_SP)
#define BLEP_RNS 7

typedef struct blep_t
{
    int32_t index, samplesLeft;
    float buffer[BLEP_RNS + 1], lastValue;
} blep_t;

#define LERP(x, y, z) ((x) + ((y) - (x)) * (z))
#define CLAMP(x, low, high) (((x) > (high)) ? (high) : (((x) < (low)) ? (low) : (x)))
#define CLAMP16(i) \
{ \
    if ((int16_t)(i) != i) \
        i = 0x7FFF ^ (i >> 31); \
}

typedef struct
{
    int32_t Note, Fixed, Waveform, FX[2], FXParam[2];
} AHXPListEntry;

typedef struct
{
    int32_t Speed, Length;
    AHXPListEntry *Entries;
} AHXPList;

typedef struct
{
    int32_t aFrames, aVolume;
    int32_t dFrames, dVolume;
    int32_t sFrames;
    int32_t rFrames, rVolume;
} AHXEnvelope;

typedef struct
{
    char* Name;
    int32_t Volume, WaveLength;
    int32_t FilterLowerLimit, FilterUpperLimit, FilterSpeed;
    int32_t SquareLowerLimit, SquareUpperLimit, SquareSpeed;
    int32_t VibratoDelay, VibratoDepth, VibratoSpeed;
    int32_t HardCutRelease, HardCutReleaseFrames;
    AHXEnvelope Envelope;
    AHXPList PList;
} AHXInstrument;

typedef struct
{
    int32_t Track[AMIGA_VOICES], Transpose[AMIGA_VOICES];
} AHXPosition;

typedef struct
{
    int32_t Note, Instrument, FX, FXParam;
} AHXStep;

typedef struct
{
    int8_t VoiceBuffer[0x280], SquareTempBuffer[0x80], *AudioPointer, *AudioSource;
    int32_t InstrPeriod, TrackPeriod, VibratoPeriod;
    int32_t VoiceVolume, VoicePeriod;
    int32_t Track, Transpose;
    int32_t NextTrack, NextTranspose;
    int32_t ADSRVolume;
    int32_t NoteMaxVolume, PerfSubVolume, TrackMasterVolume;
    int32_t NewWaveform, Waveform, PlantSquare, PlantPeriod, IgnoreSquare;
    int32_t TrackOn, FixedNote;
    int32_t VolumeSlideUp, VolumeSlideDown;
    int32_t HardCut, HardCutRelease, HardCutReleaseF;
    int32_t PeriodSlideSpeed, PeriodSlidePeriod, PeriodSlideLimit, PeriodSlideOn, PeriodSlideWithLimit;
    int32_t PeriodPerfSlideSpeed, PeriodPerfSlidePeriod, PeriodPerfSlideOn;
    int32_t VibratoDelay, VibratoCurrent, VibratoDepth, VibratoSpeed;
    int32_t SquareOn, SquareInit, SquareWait, SquareLowerLimit, SquareUpperLimit, SquarePos, SquareSign, SquareSlidingIn, SquareReverse;
    int32_t FilterOn, FilterInit, FilterWait, FilterLowerLimit, FilterUpperLimit, FilterPos, FilterSign, FilterSpeed, FilterSlidingIn, IgnoreFilter;
    int32_t PerfCurrent, PerfSpeed, PerfWait;
    int32_t WaveLength;
    int32_t NoteDelayWait, NoteDelayOn, NoteCutWait, NoteCutOn;
    int32_t AudioPeriod, AudioVolume;
    AHXEnvelope ADSR;
    AHXInstrument *Instrument;
    AHXPList *PerfList;
} AHXVoice;

static struct
{
    char *Name;
    int32_t *Subsongs;
    int32_t Restart, PositionNr, TrackLength, TrackNr, InstrumentNr, SubsongNr;
    int32_t Revision, SpeedMultiplier;
    AHXPosition *Positions;
    AHXStep **Tracks;
    AHXInstrument *Instruments;
} Song;

static struct
{
    int8_t LowPasses[(0xFC + 0xFC + 0x80 * 0x1F + 0x80 + 3 * 0x280) * FILTER_ITERATIONS];
    int8_t Triangle04[0x04], Triangle08[0x08], Triangle10[0x10], Triangle20[0x20], Triangle40[0x40], Triangle80[0x80];
    int8_t Sawtooth04[0x04], Sawtooth08[0x08], Sawtooth10[0x10], Sawtooth20[0x20], Sawtooth40[0x40], Sawtooth80[0x80];
    int8_t Squares[0x80 * 32];
    int8_t WhiteNoiseBig[0x280 * 3];
    int8_t HighPasses[(0xFC + 0xFC + 0x80 * 0x1F + 0x80 + 3 * 0x280) * FILTER_ITERATIONS];
} Waves;

typedef struct paulaVoice_t
{
    const int8_t *SRC_DAT, *DMA_DAT;
    uint32_t SRC_LEN, DMA_LEN, DMA_POS;
    float SRC_VOL, DELTA, FRAC, LASTDELTA, LASTFRAC, PANL, PANR;
} paulaVoice_t;

#if defined(USE_HIGHPASS)
typedef struct lossyIntegrator_t
{
    float buffer[2], coeff[2];
} lossyIntegrator_t;
#endif

static const int16_t midFilterTable[31 * (6 + 6 + 0x20 + 1)];
static const int16_t lowFilterTable[31 * (6 + 6 + 0x20 + 1)];
static int8_t *WaveformTab[4], SongLoaded;
static int32_t StepWaitFrames, GetNewPosition, SongEndReached;
static int32_t PatternBreak;
static int32_t MainVolume;
static int32_t Playing, Tempo;
static int32_t PosNr, PosJump;
static int32_t NoteNr, PosJumpNote;
static int32_t WNRandom;
static int32_t soundBufferSize;
static float f_outputFreq;
static float *mixerBufferL = NULL, *mixerBufferR = NULL;
static int8_t *winmmBuffer = NULL;
static int32_t samplesLeft;
static int32_t samplesPerFrame;
static volatile int8_t mixingMutex, isMixing;
static paulaVoice_t AUD[AMIGA_VOICES];
static AHXVoice Voices[AMIGA_VOICES];
#ifdef USE_BLEP
static blep_t blep[4], blepVol[4];
#endif
#ifdef USE_HIGHPASS
static lossyIntegrator_t filterHi;
#endif

const int8_t Offsets[6] = { 0x00, 0x04, 0x04+0x08, 0x04+0x08+0x10, 0x04+0x08+0x10+0x20, 0x04+0x08+0x10+0x20+0x40 };

const uint16_t waveformLen[NUM_WAVEFORMS] =
{
    0x04, 0x08, 0x10, 0x20, 0x40, 0x80, // sawtooth
    0x04, 0x08, 0x10, 0x20, 0x40, 0x80, // triangle

    // squarewave
    0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
    0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
    0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
    0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,

    0x280 * 3 // whitenoise
};

static const int16_t VibratoTable[64] =
{
    0,24,49,74,97,120,141,161,180,197,212,224,235,244,250,253,255,
    253,250,244,235,224,212,197,180,161,141,120,97,74,49,24,
    0,-24,-49,-74,-97,-120,-141,-161,-180,-197,-212,-224,-235,-244,-250,-253,-255,
    -253,-250,-244,-235,-224,-212,-197,-180,-161,-141,-120,-97,-74,-49,-24
};

static const int16_t PeriodTable[1 + (5 * 12)] =
{
    0x0000, 0x0D60, 0x0CA0, 0x0BE8, 0x0B40, 0x0A98, 0x0A00, 0x0970,
    0x08E8, 0x0868, 0x07F0, 0x0780, 0x0714, 0x06B0, 0x0650, 0x05F4,
    0x05A0, 0x054C, 0x0500, 0x04B8, 0x0474, 0x0434, 0x03F8, 0x03C0,
    0x038A, 0x0358, 0x0328, 0x02FA, 0x02D0, 0x02A6, 0x0280, 0x025C,
    0x023A, 0x021A, 0x01FC, 0x01E0, 0x01C5, 0x01AC, 0x0194, 0x017D,
    0x0168, 0x0153, 0x0140, 0x012E, 0x011D, 0x010D, 0x00FE, 0x00F0,
    0x00E2, 0x00D6, 0x00CA, 0x00BE, 0x00B4, 0x00AA, 0x00A0, 0x0097,
    0x008F, 0x0087, 0x007F, 0x0078, 0x0071
};

#ifdef USE_BLEP
static const uint32_t blepData[48] =
{
    0x3F7FE1F1, 0x3F7FD548, 0x3F7FD6A3, 0x3F7FD4E3,
    0x3F7FAD85, 0x3F7F2152, 0x3F7DBFAE, 0x3F7ACCDF,
    0x3F752F1E, 0x3F6B7384, 0x3F5BFBCB, 0x3F455CF2,
    0x3F26E524, 0x3F0128C4, 0x3EACC7DC, 0x3E29E86B,
    0x3C1C1D29, 0xBDE4BBE6, 0xBE3AAE04, 0xBE48DEDD,
    0xBE22AD7E, 0xBDB2309A, 0xBB82B620, 0x3D881411,
    0x3DDADBF3, 0x3DE2C81D, 0x3DAAA01F, 0x3D1E769A,
    0xBBC116D7, 0xBD1402E8, 0xBD38A069, 0xBD0C53BB,
    0xBC3FFB8C, 0x3C465FD2, 0x3CEA5764, 0x3D0A51D6,
    0x3CEAE2D5, 0x3C92AC5A, 0x3BE4CBF7, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000
};
#endif

static void GenerateWaves(void);
static void ProcessStep(int32_t v);
static void ProcessFrame(int32_t v);
static void SetAudio(int32_t v);
static void PListCommandParse(int32_t v, int32_t FX, int32_t FXParam);
static void PlayIRQ(void);

char *ahxplay_GetSongName(void)
{
    if (SongLoaded)
        return (Song.Name);

    return (NULL);
}

static void PaulaRestartDMA(uint8_t i)
{
    AUD[i].FRAC    = 0.0f;
    AUD[i].DMA_POS = 0;
    AUD[i].DMA_DAT = AUD[i].SRC_DAT;
    AUD[i].DMA_LEN = AUD[i].SRC_LEN;
}

static void PaulaSetPeriod(uint8_t i, uint16_t period)
{
    if (period == 0)
    {
        AUD[i].DELTA = 0.0f;
    }
    else
    {
        /* confirmed behavior on real Amiga */
        if (period < 113)
            period = 113;

        AUD[i].DELTA = (3546895.0f / period) / f_outputFreq;
    }

    if (AUD[i].LASTDELTA == 0.0f)
        AUD[i].LASTDELTA = AUD[i].DELTA;
}

static void PaulaSetVolume(uint8_t i, uint16_t vol)
{
    vol &= 0x007F;
    if (vol > 0x40)
        vol = 0x40;

    AUD[i].SRC_VOL = vol * (1.0f / 64.0f);
}

static void PaulaSetLength(uint8_t i, uint16_t len)
{
    AUD[i].SRC_LEN = 2 * len;
}

static void PaulaSetData(uint8_t i, const int8_t *src)
{
    AUD[i].SRC_DAT = src;
}

#if defined(USE_HIGHPASS)
static void calcCoeffLossyIntegrator(float sr, float hz, lossyIntegrator_t *filter)
{
    filter->coeff[0] = tanf((3.1415927f * hz) / sr);
    filter->coeff[1] = 1.0f / (1.0f + filter->coeff[0]);
}

static void clearLossyIntegrator(lossyIntegrator_t *filter)
{
    filter->buffer[0] = 0.0f;
    filter->buffer[1] = 0.0f;
}

static inline void lossyIntegrator(lossyIntegrator_t *filter, float *in, float *out)
{
    float output;

    /* left channel low-pass */
    output = (filter->coeff[0] * in[0] + filter->buffer[0]) * filter->coeff[1];
    filter->buffer[0] = filter->coeff[0] * (in[0] - output) + output + 1e-10f;
    out[0] = output;

    /* right channel low-pass */
    output = (filter->coeff[0] * in[1] + filter->buffer[1]) * filter->coeff[1];
    filter->buffer[1] = filter->coeff[0] * (in[1] - output) + output + 1e-10f;
    out[1] = output;
}

static inline void lossyIntegratorHighPass(lossyIntegrator_t *filter, float *in, float *out)
{
    float low[2];

    lossyIntegrator(filter, in, low);

    out[0] = in[0] - low[0];
    out[1] = in[1] - low[1];
}
#endif

static void freeSong(void)
{
    int32_t i;

    if (Song.Subsongs != NULL)
    {
        free(Song.Subsongs);
        Song.Subsongs = NULL;
    }

    if (Song.Positions != NULL)
    {
        free(Song.Positions);
        Song.Positions = NULL;
    }

    if (Song.Tracks != NULL)
    {
        free(Song.Tracks);
        Song.Tracks = NULL;
    }

    if (Song.Instruments != NULL)
    {
        for (i = 1; i < Song.InstrumentNr + 1; ++i)
        {
            if (Song.Instruments[i].PList.Entries != NULL)
            {
                free(Song.Instruments[i].PList.Entries);
                Song.Instruments[i].PList.Entries = NULL;
            }
        }

        free(Song.Instruments);
        Song.Instruments = NULL;
    }
}

#ifdef USE_BLEP
static inline void blepAdd(blep_t *b, float offset, float amplitude)
{
    int8_t n;
    uint32_t i;
    const float *blepSrc;
    float f;

    if ((offset < 0.0f) || (offset > 1.0f))
        return;

    i  = (uint32_t)(offset * BLEP_SP);
    blepSrc = (const float *)(blepData) + i + BLEP_OS;

    f = (offset * BLEP_SP) - i;
    i = b->index;

    n = BLEP_NS;
    while (n--)
    {
        b->buffer[i] += (amplitude * LERP(blepSrc[0], blepSrc[1], f));

        i++;
        i &= BLEP_RNS;

        blepSrc += BLEP_SP;
    }

    b->samplesLeft = BLEP_NS;
}

static inline float blepRun(blep_t *b)
{
    float blepOutput;

    blepOutput = b->buffer[b->index];
    b->buffer[b->index] = 0.0f;

    b->index++;
    b->index &= BLEP_RNS;

    b->samplesLeft--;

    return (blepOutput);
}
#endif

static void PlayIRQ(void) /* main music interrupt */
{
    int32_t NextPos;
    uint8_t i;
    AHXVoice *voice;

    if (!Playing)
        return;

    if (StepWaitFrames <= 0)
    {
        if (GetNewPosition)
        {
            NextPos = ((PosNr + 1) == Song.PositionNr) ? 0 : (PosNr + 1);

            for (i = 0; i < AMIGA_VOICES; ++i)
            {
                voice = &Voices[i];

                voice->Track         = Song.Positions[PosNr].Track[i];
                voice->Transpose     = Song.Positions[PosNr].Transpose[i];
                voice->NextTrack     = Song.Positions[NextPos].Track[i];
                voice->NextTranspose = Song.Positions[NextPos].Transpose[i];
            }

            GetNewPosition = false;
        }

        for (i = 0; i < AMIGA_VOICES; ++i)
            ProcessStep(i);

        StepWaitFrames = Tempo;
    }

    for (i = 0; i < AMIGA_VOICES; ++i)
        ProcessFrame(i);

    if ((Tempo > 0) && (--StepWaitFrames <= 0))
    {
        if (!PatternBreak)
        {
            if (++NoteNr >= Song.TrackLength)
            {
                PosJump      = PosNr + 1;
                PosJumpNote  = 0;
                PatternBreak = true;
            }
        }

        if (PatternBreak)
        {
            PatternBreak = false;

            PosNr  = PosJump;
            NoteNr = PosJumpNote;

            PosJumpNote = 0;
            PosJump     = 0;

            if (PosNr == Song.PositionNr)
            {
                SongEndReached = true;
                PosNr = Song.Restart;
            }

            GetNewPosition = true;
        }
    }

    for (i = 0; i < AMIGA_VOICES; ++i)
        SetAudio(i);
}

static void CalcADSR(AHXVoice *v)
{
    int32_t af, df, rf;

    af = v->ADSR.aFrames = v->Instrument->Envelope.aFrames;
    df = v->ADSR.dFrames = v->Instrument->Envelope.dFrames;
         v->ADSR.sFrames = v->Instrument->Envelope.sFrames;
    rf = v->ADSR.rFrames = v->Instrument->Envelope.rFrames;

    if (af == 0) af = 1;
    if (df == 0) df = 1;
    if (rf == 0) rf = 0;

    v->ADSR.aVolume =  (v->Instrument->Envelope.aVolume * 256) / af;
    v->ADSR.dVolume = ((v->Instrument->Envelope.dVolume - v->Instrument->Envelope.aVolume) * 256) / df;
    v->ADSR.rVolume = ((v->Instrument->Envelope.rVolume  -v->Instrument->Envelope.dVolume) * 256) / rf;
}

static void ProcessStep(int32_t v)
{
    int32_t i, Note, Instrument, FX, FXParam, t, Neue, Alte, SquareLower, SquareUpper, d3, d4, d6;
    AHXVoice *voice;
    AHXStep *step;

    voice = &Voices[v];

    if (!voice->TrackOn)
        return;

    voice->VolumeSlideUp = voice->VolumeSlideDown = 0;

    step = &Song.Tracks[Song.Positions[PosNr].Track[v]][NoteNr];

    Note       = step->Note;
    Instrument = step->Instrument;
    FX         = step->FX;
    FXParam    = step->FXParam;

    switch (FX)
    {
        case 0x00: // Position Jump HI
        {
            if (((FXParam & 0x0F) > 0) && ((FXParam & 0x0F) <= 9))
                PosJump = FXParam & 0x0F;
        }
        break;

        case 0x05: // Volume Slide + Tone Portamento
        case 0x0A: // Volume Slide
        {
            voice->VolumeSlideDown = FXParam & 0x0F;
            voice->VolumeSlideUp   = FXParam >> 4;
        }
        break;

        case 0x0B: // Position Jump
        {
            PosJump = (PosJump * 100) + (FXParam & 0x0F) + ((FXParam >> 4) * 10);
            PatternBreak = true;
        }
        break;

        case 0x0D: // Patternbreak
        {
            PosJump = PosNr + 1;

            PosJumpNote = (FXParam & 0x0F) + (FXParam >> 4) * 10;
            if (PosJumpNote > Song.TrackLength)
                PosJumpNote = 0;

            PatternBreak = true;
        }
        break;

        case 0x0E: // Enhanced commands
        {
            switch (FXParam >> 4)
            {
                case 0xC: // Note Cut
                {
                    if ((FXParam & 0x0f) < Tempo)
                    {
                        voice->NoteCutWait = FXParam & 0x0F;
                        if(voice->NoteCutWait)
                        {
                            voice->NoteCutOn      = true;
                            voice->HardCutRelease = false;
                        }
                    }
                }
                break;

                case 0xD: // Note Delay
                {
                    if (voice->NoteDelayOn)
                    {
                        voice->NoteDelayOn = false;
                    }
                    else
                    {
                        if ((FXParam & 0x0f) < Tempo)
                        {
                            voice->NoteDelayWait = FXParam & 0x0F;
                            if (voice->NoteDelayWait)
                            {
                                voice->NoteDelayOn = true;
                                return;
                            }
                        }
                    }
                }
                break;
            }
        }
        break;

        case 0x0F: // Speed
            Tempo = FXParam;
        break;
    }

    if (Instrument && (Instrument <= Song.InstrumentNr))
    {
        voice->PerfSubVolume    = 0x40;
        voice->PeriodSlideSpeed = voice->PeriodSlidePeriod = voice->PeriodSlideLimit = 0;
        voice->ADSRVolume       = 0;
        voice->Instrument       = &Song.Instruments[Instrument];

        CalcADSR(voice);

        voice->WaveLength     = voice->Instrument->WaveLength;
        voice->NoteMaxVolume  = voice->Instrument->Volume;
        voice->VibratoCurrent = 0;
        voice->VibratoDelay   = voice->Instrument->VibratoDelay;
        voice->VibratoDepth   = voice->Instrument->VibratoDepth;
        voice->VibratoSpeed   = voice->Instrument->VibratoSpeed;
        voice->VibratoPeriod  = 0;
        voice->HardCutRelease = voice->Instrument->HardCutRelease;
        voice->HardCut        = voice->Instrument->HardCutReleaseFrames;
        voice->IgnoreSquare   = voice->SquareSlidingIn = 0;
        voice->SquareWait     = voice->SquareOn = 0;

        SquareLower = voice->Instrument->SquareLowerLimit >> (5 - voice->WaveLength);
        SquareUpper = voice->Instrument->SquareUpperLimit >> (5 - voice->WaveLength);

        if (SquareUpper < SquareLower)
        {
            t = SquareUpper;
            SquareUpper = SquareLower;
            SquareLower = t;
        }

        voice->SquareUpperLimit = SquareUpper;
        voice->SquareLowerLimit = SquareLower;
        voice->IgnoreFilter     = voice->FilterWait = voice->FilterOn = 0;
        voice->FilterSlidingIn  = false;

        d6 = voice->Instrument->FilterSpeed;
        d3 = voice->Instrument->FilterLowerLimit;
        d4 = voice->Instrument->FilterUpperLimit;

        if (d3 & 0x80) d6 |= 0x20;
        if (d4 & 0x80) d6 |= 0x40;

        voice->FilterSpeed = d6;

        d3 &= ~0x80;
        d4 &= ~0x80;

        if (d3 > d4)
        {
            t  = d3;
            d3 = d4;
            d4 = t;
        }

        voice->FilterUpperLimit = d4;
        voice->FilterLowerLimit = d3;

        voice->FilterPos = 32;
        voice->PerfWait  = voice->PerfCurrent = 0;
        voice->PerfSpeed = voice->Instrument->PList.Speed;
        voice->PerfList  = &voice->Instrument->PList;
    }

    voice->PeriodSlideOn = false;

    FX &= 0x0F;

    switch (FX)
    {
        case 0x09: // Set Squarewave-Offset
        {
            voice->SquarePos    = FXParam >> (5 - voice->WaveLength);
            //voice->PlantSquare  = true;   (bug ?)
            voice->IgnoreSquare = true;
        }
        break;

        case 0x03: // Tone Portamento (Period Slide Up/Down w/ Limit)
            if (FXParam != 0)
                voice->PeriodSlideSpeed = FXParam; // fall through
        case 0x05: // Tone Portamento + Volume Slide
        {
            if (Note)
            {
                Neue = PeriodTable[Note];
                Alte = PeriodTable[voice->TrackPeriod];
                Alte -= Neue;
                Neue = Alte + voice->PeriodSlidePeriod;
                if (Neue)
                    voice->PeriodSlideLimit = -Alte;
            }

            voice->PeriodSlideOn        = true;
            voice->PeriodSlideWithLimit = true;

            Note = 0;
        }
        break;
    }

    if (Note)
    {
        voice->TrackPeriod = Note;
        voice->PlantPeriod = true;
    }

    switch (FX)
    {
        case 0x01: // Portamento up (Period slide down)
        {
            voice->PeriodSlideSpeed     = -FXParam;
            voice->PeriodSlideOn        = true;
            voice->PeriodSlideWithLimit = false;
        }
        break;

        case 0x02: // Portamento down (Period slide up)
        {
            voice->PeriodSlideSpeed     = FXParam;
            voice->PeriodSlideOn        = true;
            voice->PeriodSlideWithLimit = false;
        }
        break;

        case 0x04: // Filter override
        {
            if ((FXParam == 0) || (FXParam == 0x40))
                break;

            if (FXParam < 0x40)
            {
                voice->IgnoreFilter = FXParam;
                break;
            }

            if (FXParam > 0x7F)
                break;

            voice->FilterPos = FXParam - 0x40;
        }
        break;

        case 0x0C: // Volume
        {
            FXParam &= 0xFF;
            if (FXParam <= 0x40)
            {
                voice->NoteMaxVolume = FXParam;
            }
            else
            {
                FXParam -= 0x50;
                if (FXParam <= 0x40)
                {
                    for (i = 0; i < AMIGA_VOICES; ++i)
                        Voices[i].TrackMasterVolume = FXParam;
                }
                else
                {
                    FXParam -= 0xa0 - 0x50;
                    if (FXParam <= 0x40)
                        voice->TrackMasterVolume = FXParam;
                }
            }
        }
        break;

        case 0x0E: // Extended commands
        {
            switch (FXParam >> 4)
            {
                case 0x1: // Fineslide up (Period fineslide down)
                {
                    voice->PeriodSlidePeriod -= (FXParam & 0x0F);
                    voice->PlantPeriod = true;
                }
                break;

                case 0x2: // Fineslide down (Period fineslide up)
                {
                    voice->PeriodSlidePeriod += (FXParam & 0x0F);
                    voice->PlantPeriod = true;
                }
                break;

                case 0x4: // Vibrato control
                    voice->VibratoDepth = FXParam & 0x0F;
                break;

                case 0xA: // Finevolume up
                {
                    voice->NoteMaxVolume += (FXParam & 0x0F);
                    if (voice->NoteMaxVolume > 0x40)
                        voice->NoteMaxVolume = 0x40;
                }
                break;

                case 0xB: // Finevolume down
                {
                    voice->NoteMaxVolume -= (FXParam & 0x0F);
                    if (voice->NoteMaxVolume < 0)
                        voice->NoteMaxVolume = 0;
                }
                break;
            }
        }
        break;
    }
}

static void ProcessFrame(int32_t v)
{
    int8_t *SquarePtr, *AudioSource;
    int32_t i, d0, d1, d2, d3, NextInstrument, Cur, FMax, Delta, X;
    AHXVoice *voice;

    voice = &Voices[v];

    if (!voice->TrackOn)
        return;

    if (voice->NoteDelayOn)
    {
        if (voice->NoteDelayWait <= 0)
            ProcessStep(v);
        else
            voice->NoteDelayWait--;
    }

    if (voice->HardCut)
    {
        if ((NoteNr + 1) < Song.TrackLength)
            NextInstrument = Song.Tracks[voice->Track][NoteNr+1].Instrument;
        else
            NextInstrument = Song.Tracks[voice->NextTrack][0].Instrument;

        if (NextInstrument)
        {
            d1 = Tempo - voice->HardCut;
            if (d1 < 0)
                d1 = 0;

            if (!voice->NoteCutOn)
            {
                voice->NoteCutOn       = true;
                voice->NoteCutWait     = d1;
                voice->HardCutReleaseF = -(d1 - Tempo);
            }
            else
            {
                voice->HardCut = 0;
            }
        }
    }

    if (voice->NoteCutOn)
    {
        if (voice->NoteCutWait <= 0)
        {
            voice->NoteCutOn = false;

            if (voice->HardCutRelease)
            {
                voice->ADSR.rVolume = -(voice->ADSRVolume - (voice->Instrument->Envelope.rVolume << 8)) / voice->HardCutReleaseF;
                voice->ADSR.rFrames = voice->HardCutReleaseF;
                voice->ADSR.aFrames = voice->ADSR.dFrames = voice->ADSR.sFrames = 0;
            }
            else
            {
                voice->NoteMaxVolume = 0;
            }
        }
        else
        {
            voice->NoteCutWait--;
        }
    }

    if (voice->ADSR.aFrames)
    {
        voice->ADSRVolume += voice->ADSR.aVolume; // Delta
        if (--voice->ADSR.aFrames <= 0)
              voice->ADSRVolume = voice->Instrument->Envelope.aVolume << 8;

    }
    else if (voice->ADSR.dFrames)
    {
        voice->ADSRVolume += voice->ADSR.dVolume; // Delta
        if (--voice->ADSR.dFrames <= 0)
              voice->ADSRVolume = voice->Instrument->Envelope.dVolume << 8;
    }
    else if (voice->ADSR.sFrames)
    {
        voice->ADSR.sFrames--;
    }
    else if (voice->ADSR.rFrames)
    {
        voice->ADSRVolume += voice->ADSR.rVolume; // Delta
        if (--voice->ADSR.rFrames <= 0)
              voice->ADSRVolume = voice->Instrument->Envelope.rVolume << 8;
    }

    voice->NoteMaxVolume = voice->NoteMaxVolume + voice->VolumeSlideUp - voice->VolumeSlideDown;
    voice->NoteMaxVolume = CLAMP(voice->NoteMaxVolume, 0x00, 0x40);

    if (voice->PeriodSlideOn)
    {
        if (voice->PeriodSlideWithLimit)
        {
            d0 = voice->PeriodSlidePeriod - voice->PeriodSlideLimit;
            d2 = voice->PeriodSlideSpeed;

            if (d0 > 0)
                d2 = -d2;

            if (d0)
            {
                d3 = (d0 + d2) ^ d0;
                if (d3 >= 0)
                    d0 = voice->PeriodSlidePeriod + d2;
                else
                    d0 = voice->PeriodSlideLimit;

                voice->PeriodSlidePeriod = d0;
                voice->PlantPeriod       = true;
            }
        }
        else
        {
            voice->PeriodSlidePeriod += voice->PeriodSlideSpeed;
            voice->PlantPeriod = true;
        }
    }

    if (voice->VibratoDepth)
    {
        if (voice->VibratoDelay <= 0)
        {
            voice->VibratoPeriod  = (VibratoTable[voice->VibratoCurrent] * voice->VibratoDepth) >> 7;
            voice->PlantPeriod    = true;
            voice->VibratoCurrent = (voice->VibratoCurrent + voice->VibratoSpeed) & 0x3F;
        }
        else
        {
            voice->VibratoDelay--;
        }
    }

    if (voice->PerfList != NULL)
    {
        if (voice->Instrument && (voice->PerfCurrent < voice->Instrument->PList.Length))
        {
            if (--voice->PerfWait <= 0)
            {
                Cur = voice->PerfCurrent++;
                voice->PerfWait = voice->PerfSpeed;

                if (voice->PerfList->Entries[Cur].Waveform)
                {
                    voice->Waveform             = voice->PerfList->Entries[Cur].Waveform - 1;
                    voice->NewWaveform          = true;
                    voice->PeriodPerfSlideSpeed = voice->PeriodPerfSlidePeriod = 0;
                }

                voice->PeriodPerfSlideOn = false;
                for (i = 0; i < 2; ++i)
                    PListCommandParse(v, voice->PerfList->Entries[Cur].FX[i] & 0xFF, voice->PerfList->Entries[Cur].FXParam[i] & 0xFF);

                if (voice->PerfList->Entries[Cur].Note)
                {
                    voice->InstrPeriod = voice->PerfList->Entries[Cur].Note;
                    voice->PlantPeriod = true;
                    voice->FixedNote   = voice->PerfList->Entries[Cur].Fixed;
                }
            }
        }
        else
        {
            if (voice->PerfWait)
                voice->PerfWait--;
            else
                voice->PeriodPerfSlideSpeed = 0;
        }
    }

    if (voice->PeriodPerfSlideOn)
    {
        voice->PeriodPerfSlidePeriod -= voice->PeriodPerfSlideSpeed;
        if (voice->PeriodPerfSlidePeriod)
            voice->PlantPeriod = true;
    }

    if ((voice->Waveform == (3 - 1)) && voice->SquareOn)
    {
        if (--voice->SquareWait <= 0)
        {
            d1 = voice->SquareLowerLimit;
            d2 = voice->SquareUpperLimit;
            d3 = voice->SquarePos;

            if (voice->SquareInit)
            {
                voice->SquareInit = false;

                if (d3 <= d1)
                {
                    voice->SquareSlidingIn = true;
                    voice->SquareSign      = 1;
                }
                else if (d3 >= d2)
                {
                    voice->SquareSlidingIn = true;
                    voice->SquareSign      = -1;
                }
            }

            if ((d1 == d3) || (d2 == d3))
            {
                if (voice->SquareSlidingIn)
                    voice->SquareSlidingIn = false;
                else
                    voice->SquareSign = -voice->SquareSign;
            }

            d3 += voice->SquareSign;

            voice->SquarePos   = d3;
            voice->PlantSquare = true;
            voice->SquareWait  = voice->Instrument->SquareSpeed;
        }
    }

    if (voice->FilterOn && (--voice->FilterWait <= 0))
    {
        d1 = voice->FilterLowerLimit;
        d2 = voice->FilterUpperLimit;
        d3 = voice->FilterPos;

        if (voice->FilterInit)
        {
            voice->FilterInit = false;

            if (d3 <= d1)
            {
                voice->FilterSlidingIn = true;
                voice->FilterSign      = 1;
            }
            else if (d3 >= d2)
            {
                voice->FilterSlidingIn = true;
                voice->FilterSign      = -1;
            }
        }

        FMax = (voice->FilterSpeed < 3) ? (5 - voice->FilterSpeed) : 1;
        for (i = 0; i < FMax; ++i)
        {
            if ((d1 == d3) || (d2 == d3))
            {
                if (voice->FilterSlidingIn)
                    voice->FilterSlidingIn = false;
                else
                    voice->FilterSign = -voice->FilterSign;
            }

            d3 += voice->FilterSign;
        }

        d3 = CLAMP(d3, 1, 63);

        voice->FilterPos   = d3;
        voice->NewWaveform = true;

        voice->FilterWait = voice->FilterSpeed - 3;
        if (voice->FilterWait < 1)
            voice->FilterWait = 1;
    }

    if ((voice->Waveform == (3 - 1)) || voice->PlantSquare)
    {
        SquarePtr = &Waves.Squares[(voice->FilterPos - 0x20) * (0xFC + 0xFC + 0x80 * 0x1F + 0x80 + 0x280 * 3)];

        X = voice->SquarePos << (5 - voice->WaveLength);
        if (X > 0x20)
        {
            X = 0x40 - X;
            voice->SquareReverse = true;
        }

        if (X > 0)
            SquarePtr += ((X - 1) << 7);

        Delta = 32 >> voice->WaveLength;

        WaveformTab[2] = voice->SquareTempBuffer;
        for (i = 0; i < ((1 << voice->WaveLength) * 4); ++i)
        {
            voice->SquareTempBuffer[i] = *SquarePtr;
            SquarePtr += Delta;
        }

        voice->NewWaveform = true;
        voice->Waveform    = 3 - 1;
        voice->PlantSquare = false;
    }

    if (voice->Waveform == (4 - 1))
        voice->NewWaveform = true;

    if (voice->NewWaveform)
    {
        AudioSource = WaveformTab[voice->Waveform];
        if (voice->Waveform != (3 - 1))
            AudioSource += (voice->FilterPos - 0x20) * (0xFC + 0xFC + 0x80 * 0x1F + 0x80 + 0x280 * 3);

        if (voice->Waveform < (3 - 1))
            AudioSource += Offsets[voice->WaveLength];

        if (voice->Waveform == (4 - 1))
        {
            AudioSource += (WNRandom & (2 * 0x280 - 1)) & ~1;
            WNRandom    += 2239384;
            WNRandom     = ((((WNRandom >> 8) | (WNRandom << 24)) + 782323) ^ 75) - 6735;
        }

        voice->AudioSource = AudioSource;
    }

    voice->AudioPeriod = voice->InstrPeriod;
    if (!voice->FixedNote)
        voice->AudioPeriod += (voice->Transpose + voice->TrackPeriod - 1);

    voice->AudioPeriod = CLAMP(voice->AudioPeriod, 0, 5 * 12);
    voice->AudioPeriod = PeriodTable[voice->AudioPeriod];

    if (!voice->FixedNote)
        voice->AudioPeriod += voice->PeriodSlidePeriod;

    voice->AudioPeriod += (voice->PeriodPerfSlidePeriod + voice->VibratoPeriod);
    voice->AudioPeriod = CLAMP(voice->AudioPeriod, 0x0071, 0x0D60);

    voice->AudioVolume = ((((((((voice->ADSRVolume >> 8) * voice->NoteMaxVolume) >> 6) * voice->PerfSubVolume) >> 6) * voice->TrackMasterVolume) >> 6) * MainVolume) >> 6;
}

static void SetAudio(int32_t v)
{
    int32_t i, WaveLoops;

    if (!Voices[v].TrackOn)
    {
        PaulaSetVolume(v, 0);
        return;
    }

    if (Voices[v].PlantPeriod)
    {
        Voices[v].PlantPeriod = false;
        PaulaSetPeriod(v, Voices[v].AudioPeriod);
    }

    if (Voices[v].NewWaveform)
    {
        Voices[v].NewWaveform = false;

        if (Voices[v].Waveform == (4 - 1))
        {
            memcpy(Voices[v].VoiceBuffer, Voices[v].AudioSource, 0x280);
        }
        else
        {
            WaveLoops = (1 << (5 - Voices[v].WaveLength)) * 5;
            for (i = 0; i < WaveLoops; ++i)
            {
                memcpy(&Voices[v].VoiceBuffer[i * (sizeof (int32_t) * (1 << Voices[v].WaveLength))],
                        Voices[v].AudioSource,     sizeof (int32_t) * (1 << Voices[v].WaveLength));
            }
        }
    }

    PaulaSetVolume(v, Voices[v].AudioVolume);
}

static void PListCommandParse(int32_t v, int32_t FX, int32_t FXParam)
{
    switch (FX)
    {
        case 0x00:
        {
            if ((Song.Revision > 0) && (FXParam > 0) && (FXParam < 0x40))
            {
                if (Voices[v].IgnoreFilter)
                {
                    Voices[v].FilterPos    = Voices[v].IgnoreFilter;
                    Voices[v].IgnoreFilter = false;
                }
                else
                {
                    Voices[v].FilterPos = FXParam;
                }

                Voices[v].NewWaveform = true;
            }
        }
        break;

        case 0x01:
        {
            Voices[v].PeriodPerfSlideSpeed = FXParam;
            Voices[v].PeriodPerfSlideOn    = true;
        }
        break;

        case 0x02:
        {
            Voices[v].PeriodPerfSlideSpeed = -FXParam;
            Voices[v].PeriodPerfSlideOn    = true;
        }
        break;

        case 0x03: // Init Square Modulation
        {
            if (!Voices[v].IgnoreSquare)
                Voices[v].SquarePos = FXParam >> (5 - Voices[v].WaveLength);
            else
                Voices[v].IgnoreSquare = false;
        }
        break;

        case 0x04: // Start/Stop Modulation
        {
            if ((Song.Revision == 0) || (FXParam == 0))
            {
                Voices[v].SquareInit = (Voices[v].SquareOn ^= 1);
                Voices[v].SquareSign = 1;
            }
            else
            {
                if (FXParam & 0x0F)
                {
                    Voices[v].SquareInit = (Voices[v].SquareOn ^= 1);

                    Voices[v].SquareSign = 1;
                    if ((FXParam & 0x0F) == 0x0F)
                        Voices[v].SquareSign = -1;
                }

                if (FXParam & 0xF0)
                {
                    Voices[v].FilterInit = (Voices[v].FilterOn ^= 1);

                    Voices[v].FilterSign = 1;
                    if ((FXParam & 0xF0) == 0xF0)
                        Voices[v].FilterSign = -1;
                }
            }
        }
        break;

        case 0x05: // Jump to Step [xx]
            Voices[v].PerfCurrent = FXParam;
        break;

        case 0x06: // Set Volume
        {
            if (FXParam > 0x40)
            {
                if ((FXParam -= 0x50) >= 0)
                {
                    if (FXParam <= 0x40)
                    {
                        Voices[v].PerfSubVolume = FXParam;
                    }
                    else
                    {
                        if ((FXParam -= (0xA0 - 0x50)) >= 0)
                        {
                            if (FXParam <= 0x40)
                                Voices[v].TrackMasterVolume = FXParam;
                        }
                    }
                }
            }
            else
            {
                Voices[v].NoteMaxVolume = FXParam;
            }
        }
        break;

        case 0x07: // set speed
            Voices[v].PerfSpeed = Voices[v].PerfWait = FXParam;
        break;
    }
}

static inline int32_t clipShifted8(int32_t in)
{
    int16_t top;

    top = (int16_t)(in >> 16);

         if (top >  127) in =   127 << 16;
    else if (top < -128) in = -(128 << 16);

    return (in);
}

static void GenerateFilterWaveforms(int8_t *buf, int8_t *loBuf, int8_t *hiBuf)
{
    const int16_t *midTablePtr = midFilterTable;
    const int16_t *lowTablePtr = lowFilterTable;
    const int8_t *waveformPtr;
    int32_t freq, i, wv, mid, low, j, in, high, fre;

    freq = 25;
    for (i = 0; i < FILTER_ITERATIONS; ++i)
    {
        waveformPtr = buf;
        for (wv = 0; wv < NUM_WAVEFORMS; ++wv)
        {
            mid = (*midTablePtr++) << 8;
            low = (*lowTablePtr++) << 8;

            for (j = 0; j < waveformLen[wv]; ++j)
            {
                in = (*waveformPtr++) << 16;

                high = clipShifted8(in - mid - low);
                fre  = (high >> 8) * freq;
                mid  = clipShifted8(mid + fre);
                fre  = (mid  >> 8) * freq;
                low  = clipShifted8(low + fre);

                *loBuf++ = low  >> 16;
                *hiBuf++ = high >> 16;
            }
        }

        freq += 9;
    }
}

static void GenerateTriangle(int8_t *buf, int32_t len)
{
    int8_t c, *buf2;
    int32_t i, d2, d5, d1, d4, val;

    d2  = len;
    d5  = len / 4;
    d1  = 128 / d5;
    d4  = -(d2 / 2);
    val = 0;

    for (i = 0; i < d5; ++i)
    {
        *buf++ = val;
        val   += d1;
    }

    *buf++ = 0x7F;

    if (d5 != 1)
    {
        val = 128;
        for (i = 0; i < (d5 - 1); ++i)
        {
            val   -= d1;
            *buf++ = val;
        }
    }

    buf2 = buf + d4;
    for (i = 0; i < (d5 * 2); ++i)
    {
        c = *buf2++;
        if (c == 0x7F)
            c = 0x80;
        else
            c = -c;

        *buf++ = c;
    }
}

static void GenerateSquare(int8_t *buf)
{
    int32_t i, j;

    for (i = 1; i <= 32; ++i)
    {
        for (j = 0; j < ((64 - i) * 2); ++j)
            *buf++ = 0x80;

        for (j = 0; j < (i * 2); ++j)
            *buf++ = 0x7F;
    }
}

static void GenerateSawtooth(int8_t *buf, int32_t len)
{
    int8_t val;
    int32_t i;

    val = -128;
    for (i = 0; i < len; ++i)
    {
        *buf++ = val;
        val   += (256 / (len - 1));
    }
}

static void GenerateWhiteNoise(int8_t *buf, int32_t len)
{
    int8_t s;
    uint16_t ax, bx;
    int32_t i;
    uint32_t ays;

    ays = 0x41595321; // seed ("AYS!")
    for (i = 0; i < len; ++i)
    {
        s = ays;
        if (ays & 0x100)
        {
            s = 0x7f;
            if (ays & 0x8000)
                s = 0x80;
        }

        *buf++ = s;

        ays  = (ays >> 5) | (ays << 27);
        ays  = (ays & 0xFFFFFF00) | ((ays & 0xFF) ^ 0x9A);
        bx   = ays;
        ays  = (ays << 2) | (ays >> 30);
        ax   = ays;
        bx  += ax;
        ax  ^= bx;
        ays  = (ays & 0xFFFF0000) | ax;
        ays  = (ays >> 3) | (ays << 29);
    }
}

static void GenerateWaves(void)
{
    // generates bit-perfect waveforms to the original Amiga replayer

    GenerateSawtooth(Waves.Sawtooth04, 0x04);
    GenerateSawtooth(Waves.Sawtooth08, 0x08);
    GenerateSawtooth(Waves.Sawtooth10, 0x10);
    GenerateSawtooth(Waves.Sawtooth20, 0x20);
    GenerateSawtooth(Waves.Sawtooth40, 0x40);
    GenerateSawtooth(Waves.Sawtooth80, 0x80);
    GenerateTriangle(Waves.Triangle04, 0x04);
    GenerateTriangle(Waves.Triangle08, 0x08);
    GenerateTriangle(Waves.Triangle10, 0x10);
    GenerateTriangle(Waves.Triangle20, 0x20);
    GenerateTriangle(Waves.Triangle40, 0x40);
    GenerateTriangle(Waves.Triangle80, 0x80);
    GenerateSquare(Waves.Squares);
    GenerateWhiteNoise(Waves.WhiteNoiseBig, 0x280 * 3);
    GenerateFilterWaveforms(Waves.Triangle04, Waves.LowPasses, Waves.HighPasses);
}

static void mixSampleBlock(int16_t *streamOut, uint32_t numSamples)
{
    uint8_t i;
    int16_t *sndOut;
    uint16_t j;
    int32_t smpL, smpR;
    float tempSample, tempVolume, out_f[2];
    paulaVoice_t *v;
#ifdef USE_BLEP
    blep_t *bSmp, *bVol;
#endif

    memset(mixerBufferL, 0, sizeof (float) * numSamples);
    memset(mixerBufferR, 0, sizeof (float) * numSamples);

    for (i = 0; i < AMIGA_VOICES; ++i)
    {
        v = &AUD[i];

#ifdef USE_BLEP
        bSmp = &blep[i];
        bVol = &blepVol[i];
#endif
        for (j = 0; j < numSamples; ++j)
        {
            tempSample = (v->DMA_DAT == NULL) ? 0.0f : (v->DMA_DAT[v->DMA_POS] * (1.0f / 128.0f));
            tempVolume = (v->DMA_DAT == NULL) ? 0.0f : v->SRC_VOL;

#ifdef USE_BLEP
            if (tempSample != bSmp->lastValue)
            {
                if ((v->LASTDELTA > 0.0f) && (v->LASTDELTA > v->LASTFRAC))
                    blepAdd(bSmp, v->LASTFRAC / v->LASTDELTA, bSmp->lastValue - tempSample);

                bSmp->lastValue = tempSample;
            }

            if (tempVolume != bVol->lastValue)
            {
                blepAdd(bVol, 0.0f, bVol->lastValue - tempVolume);
                bVol->lastValue = tempVolume;
            }

            if (bSmp->samplesLeft) tempSample += blepRun(bSmp);
            if (bVol->samplesLeft) tempVolume += blepRun(bVol);
#endif

            tempSample *= tempVolume;

            mixerBufferL[j] += (tempSample * v->PANL);
            mixerBufferR[j] += (tempSample * v->PANR);

            v->FRAC += v->DELTA;
            if (v->FRAC >= 1.0f)
            {
                v->FRAC -= 1.0f;

                v->LASTFRAC  = v->FRAC;
                v->LASTDELTA = v->DELTA;

                if (++v->DMA_POS >= v->DMA_LEN)
                {
                    v->DMA_POS = 0;

                    /* re-fetch Paula register values now */
                    v->DMA_LEN = v->SRC_LEN;
                    v->DMA_DAT = v->SRC_DAT;
                }
            }
        }
    }

    sndOut = streamOut;
    for (j = 0; j < numSamples; ++j)
    {
        out_f[0] = mixerBufferL[j];
        out_f[1] = mixerBufferR[j];

#ifdef USE_HIGHPASS
        lossyIntegratorHighPass(&filterHi, out_f, out_f);
#endif
        /* normalize amplitude */
        out_f[0] *= (32767.0f / NORM_FACTOR);
        out_f[1] *= (32767.0f / NORM_FACTOR);

        /* truncate to signed 32-bit integer (should compile into using SSE) */
        smpL = (int32_t)(out_f[0]);
        smpR = (int32_t)(out_f[1]);

        /* clamp to 16-bit signed range (fast) */
        CLAMP16(smpL);
        CLAMP16(smpR);

        /* truncate 32-bit to signed 16-bit integer (fast) and store in output buffer */
        *sndOut++ = (int16_t)(smpL);
        *sndOut++ = (int16_t)(smpR);
    }
}

static float sinApx(float x)
{
    x = x * (2.0f - x);
    return (x * 1.09742972f + x * x * 0.31678383f);
}

static float cosApx(float x)
{
    x = (1.0f - x) * (1.0f + x);
    return (x * 1.09742972f + x * x * 0.31678383f);
}

static void calculatePans(int8_t stereoSeparation)
{
    uint8_t scaledPanPos;
    float p;

    if (stereoSeparation > 100)
        stereoSeparation = 100;

    scaledPanPos = (stereoSeparation * 128) / 100;

    p = (128 - scaledPanPos) * (1.0f / 256.0f);
    AUD[0].PANL = cosApx(p);
    AUD[0].PANR = sinApx(p);
    AUD[3].PANL = cosApx(p);
    AUD[3].PANR = sinApx(p);

    p = (128 + scaledPanPos) * (1.0f / 256.0f);
    AUD[1].PANL = cosApx(p);
    AUD[1].PANR = sinApx(p);
    AUD[2].PANL = cosApx(p);
    AUD[2].PANR = sinApx(p);
}

void ahxplay_FillAudioBuffer(int16_t *buffer, int32_t samples)
{
    int32_t samplesTodo;

    while (samples)
    {
        samplesTodo = (samples < samplesLeft) ? samples : samplesLeft;
        if (samplesTodo > soundBufferSize)
            samplesTodo = soundBufferSize;

        if (samplesTodo > 0)
        {
            mixSampleBlock(buffer, samplesTodo);
            buffer += (samplesTodo * 2);

            samples -= samplesTodo;
            samplesLeft -= samplesTodo;
        }
        else
        {
            PlayIRQ();
            samplesLeft = samplesPerFrame;
        }
    }
}

/* the following must be changed if you want to use another audio API than WinMM */

#define MIX_BUF_NUM 4
static WAVEHDR waveBlocks[MIX_BUF_NUM];
static HWAVEOUT hWaveOut;
static WAVEFORMATEX wfx;

static void CALLBACK callbackWinMM(HWAVEOUT _hWaveOut, UINT uMsg,
    DWORD_PTR dwInstance, DWORD_PTR dwParam1, DWORD_PTR dwParam2)
{
    WAVEHDR *waveBlockHeader;

    /* make compiler happy! (warning C4100) */
    (void)(dwParam2);
    (void)(dwInstance);

    if (uMsg == MM_WOM_DONE)
    {
        mixingMutex = true;

        waveBlockHeader = (WAVEHDR *)(dwParam1);
        waveOutUnprepareHeader(_hWaveOut, waveBlockHeader, sizeof (WAVEHDR));

        if (isMixing)
        {
            memcpy(waveBlockHeader->lpData, winmmBuffer, soundBufferSize);
            waveOutPrepareHeader(_hWaveOut, waveBlockHeader, sizeof (WAVEHDR));
            waveOutWrite(_hWaveOut, waveBlockHeader, sizeof (WAVEHDR));

            ahxplay_FillAudioBuffer((int16_t *)(winmmBuffer), soundBufferSize / 4);
        }

        mixingMutex = false;
    }
}

static int8_t openMixer(uint32_t _samplingFrequency, uint32_t _soundBufferSize)
{
    uint8_t i;
    MMRESULT r;

    if (!hWaveOut)
    {
        winmmBuffer = (int8_t *)(calloc(_soundBufferSize, 1));
        if (winmmBuffer == NULL)
            return (false); /* gets free'd later */

        wfx.nSamplesPerSec  = _samplingFrequency;
        wfx.wBitsPerSample  = 16;
        wfx.nChannels       = 2;
        wfx.wFormatTag      = WAVE_FORMAT_PCM;
        wfx.nBlockAlign     = (wfx.wBitsPerSample * wfx.nChannels) / 8;
        wfx.nAvgBytesPerSec = wfx.nBlockAlign * wfx.nSamplesPerSec;

        r = waveOutOpen(&hWaveOut, WAVE_MAPPER, &wfx, (DWORD_PTR)(callbackWinMM), 0L, CALLBACK_FUNCTION);
        if (r != MMSYSERR_NOERROR)
            return (false);

        for (i = 0; i < MIX_BUF_NUM; ++i)
        {
            waveBlocks[i].dwBufferLength = _soundBufferSize;

            waveBlocks[i].lpData = (LPSTR)(calloc(_soundBufferSize, 1));
            if (waveBlocks[i].lpData == NULL)
                return (false); /* gets free'd later */

            waveOutPrepareHeader(hWaveOut, &waveBlocks[i], sizeof (WAVEHDR));
            waveOutWrite(hWaveOut, &waveBlocks[i], sizeof (WAVEHDR));
        }

        return (true);
    }

    return (true);
}

static void closeMixer(void)
{
    uint8_t i;

    if (hWaveOut)
    {
        for (i = 0; i < MIX_BUF_NUM; ++i)
            waveOutUnprepareHeader(hWaveOut, &waveBlocks[i], sizeof (WAVEHDR));

        for (i = 0; i < MIX_BUF_NUM; ++i)
        {
            while (waveBlocks[i].dwFlags & WHDR_PREPARED) SleepEx(1, 1); // wait

            if (waveBlocks[i].lpData != NULL)
            {
                free(waveBlocks[i].lpData);
                waveBlocks[i].lpData = NULL;
            }
        }

        waveOutReset(hWaveOut);
        waveOutClose(hWaveOut);

        hWaveOut = 0;

        if (winmmBuffer != NULL)
        {
            free(winmmBuffer);
            winmmBuffer = NULL;
        }
    }
}

/* --------------------------------------------------------------------------- */

void ahxplay_Close(void)
{
    Playing = false;
    freeSong();

    if (isMixing)
    {
        isMixing = false;
        while (mixingMutex);
        closeMixer();
    }

    if (mixerBufferL != NULL)
    {
        free(mixerBufferL);
        mixerBufferL = NULL;
    }

    if (mixerBufferR != NULL)
    {
        free(mixerBufferR);
        mixerBufferR = NULL;
    }
}

int8_t ahxplay_Init(uint32_t outputFreq)
{
    outputFreq = CLAMP(outputFreq, 32000, 96000);

    mixerBufferL = (float *)(malloc(AUDIO_BUF_LEN * sizeof (float)));
    mixerBufferR = (float *)(malloc(AUDIO_BUF_LEN * sizeof (float)));

    if ((mixerBufferL == NULL) || (mixerBufferR == NULL))
    {
        ahxplay_Close();
        return (false);
    }

    if (!openMixer(outputFreq, AUDIO_BUF_LEN))
    {
        ahxplay_Close();
        return (false);
    }

    f_outputFreq    = (float)(outputFreq);
    soundBufferSize = AUDIO_BUF_LEN;
    samplesPerFrame = outputFreq / 50;

    calculatePans(STEREO_SEP);

#ifdef USE_HIGHPASS
    // Amiga 500 RC high-pass filter (R = 1390 ohm, C = 22uF)
    // hz = 1 / (2pi * R * C)    = ~5.2Hz
    calcCoeffLossyIntegrator(f_outputFreq, 5.2f, &filterHi);
#endif

    Song.Subsongs    = NULL;
    Song.Positions   = NULL;
    Song.Tracks      = NULL;
    Song.Instruments = NULL;

    GenerateWaves();

    WaveformTab[0] = Waves.Triangle04;
    WaveformTab[1] = Waves.Sawtooth04;
    WaveformTab[3] = Waves.WhiteNoiseBig;

    isMixing = true;
    return (true);
}

int8_t ahxplay_LoadSong(void *Buffer, int32_t Len)
{
    char *NamePtr;
    uint8_t *SongBuffer, *SBPtr;
    int32_t i, j, MaxTrack, SongLength;
    AHXInstrument *ins;
    AHXPListEntry *plist;
    AHXStep *step;
    AHXPosition *pos;

    SongLength = Len;
    SongBuffer = (uint8_t *)(Buffer);
    SBPtr      = &SongBuffer[14];

    if ((SongLength < 14) || (SongLength > 65535))
        return (false);

    if ((SongBuffer[0] != 'T') && (SongBuffer[1] != 'H') && (SongBuffer[2] != 'X'))
        return (false);

    if (SongBuffer[3] > 1)
        return (false);

    NamePtr              = (char *)(&SongBuffer[(SongBuffer[4] << 8) | SongBuffer[5]]);
    Song.Name            = _strdup(NamePtr); NamePtr += (strlen(NamePtr) + 1);
    Song.Revision        = SongBuffer[3];
    Song.SpeedMultiplier = ((SongBuffer[6] >> 5) & 3) + 1;
    Song.PositionNr      = ((SongBuffer[6] & 0x0F) << 8) | SongBuffer[7];
    Song.Restart         = (SongBuffer[8] << 8) | SongBuffer[9];
    Song.TrackLength     = SongBuffer[10];
    Song.TrackNr         = SongBuffer[11];
    Song.InstrumentNr    = SongBuffer[12];
    Song.SubsongNr       = SongBuffer[13];

    if (Song.Restart >= Song.PositionNr)
        Song.Restart  = Song.PositionNr - 1;

    Song.Subsongs = (int32_t *)(malloc(sizeof (int32_t) * Song.SubsongNr));
    if (Song.Subsongs == NULL)
    {
        freeSong();
        return (false);
    }

    for (i = 0; i < Song.SubsongNr; ++i)
    {
        if ((SBPtr - SongBuffer) > SongLength)
            return (false);

        Song.Subsongs[i] = (SBPtr[0] << 8) | SBPtr[1];
        SBPtr += 2;
    }

    Song.Positions = (AHXPosition *)(malloc(sizeof (AHXPosition) * Song.PositionNr));
    if (Song.Positions == NULL)
    {
        freeSong();
        return (false);
    }

    for (i = 0; i < Song.PositionNr; ++i)
    {
        for (j = 0; j < AMIGA_VOICES; ++j)
        {
            if ((SBPtr - SongBuffer) > SongLength)
                return (false);

            pos = &Song.Positions[i];

            pos->Track[j]     = *SBPtr++;
            pos->Transpose[j] = *((int8_t *)(SBPtr++));
        }
    }

    MaxTrack = Song.TrackNr;

    Song.Tracks = (AHXStep **)(malloc(sizeof (AHXStep) * (MaxTrack + 1)));
    if (Song.Tracks == NULL)
    {
        freeSong();
        return (false);
    }

    for (i = 0; i < MaxTrack+1; i++)
    {
        Song.Tracks[i] = (AHXStep *)(malloc(sizeof (AHXStep) * Song.TrackLength));

        if (((SongBuffer[6] & 0x80) == 0x80) && (i == 0))
        {
            memset(Song.Tracks[i], 0, Song.TrackLength * sizeof (AHXStep));
            continue;
        }

        for (j = 0; j < Song.TrackLength; ++j)
        {
            if ((SBPtr - SongBuffer) > SongLength)
                return (false);

            step = &Song.Tracks[i][j];

            step->Note       = (SBPtr[0] >> 2) & 0x3F;
            step->Instrument = ((SBPtr[0] & 0x3) << 4) | (SBPtr[1] >> 4);
            step->FX         = SBPtr[1] & 0x0F;
            step->FXParam    = SBPtr[2];

            SBPtr += 3;
        }
    }

    Song.Instruments = (AHXInstrument *)(malloc(sizeof (AHXInstrument) * (Song.InstrumentNr + 1)));
    if (Song.Instruments == NULL)
    {
        freeSong();
        return (false);
    }

    for (i = 1; i < (Song.InstrumentNr + 1); ++i)
    {
        ins = &Song.Instruments[i];

        ins->Name = _strdup(NamePtr); NamePtr += (strlen(NamePtr) + 1);
        if ((SBPtr - SongBuffer) > SongLength)
            return (false);

        ins->Volume               = SBPtr[0];
        ins->FilterSpeed          = ((SBPtr[1] >> 3) & 0x1F) | ((SBPtr[12] >> 2) & 0x20);
        ins->WaveLength           = SBPtr[1] & 7;
        ins->Envelope.aFrames     = SBPtr[2];
        ins->Envelope.aVolume     = SBPtr[3];
        ins->Envelope.dFrames     = SBPtr[4];
        ins->Envelope.dVolume     = SBPtr[5];
        ins->Envelope.sFrames     = SBPtr[6];
        ins->Envelope.rFrames     = SBPtr[7];
        ins->Envelope.rVolume     = SBPtr[8];
        ins->FilterLowerLimit     = SBPtr[12] & 0x7F;
        ins->VibratoDelay         = SBPtr[13];
        ins->HardCutReleaseFrames = (SBPtr[14] >> 4) & 7;
        ins->HardCutRelease       = (SBPtr[14] & 0x80) ? true : false;
        ins->VibratoDepth         = SBPtr[14] & 0x0F;
        ins->VibratoSpeed         = SBPtr[15];
        ins->SquareLowerLimit     = SBPtr[16];
        ins->SquareUpperLimit     = SBPtr[17];
        ins->SquareSpeed          = SBPtr[18];
        ins->FilterUpperLimit     = SBPtr[19] & 0x3F;
        ins->PList.Speed          = SBPtr[20];
        ins->PList.Length         = SBPtr[21];

        SBPtr += 22;

        ins->PList.Entries = (AHXPListEntry *)(malloc(sizeof (AHXPListEntry) * ins->PList.Length));
        if (ins->PList.Entries == NULL)
        {
            freeSong();
            return (false);
        }

        for (j = 0; j < ins->PList.Length; ++j)
        {
            if ((SBPtr - SongBuffer) > SongLength)
                return (false);

            plist = &ins->PList.Entries[j];

            plist->FX[1]      = (SBPtr[0] >> 5) & 7;
            plist->FX[0]      = (SBPtr[0] >> 2) & 7;
            plist->Waveform   = ((SBPtr[0] << 1) & 6) | (SBPtr[1] >> 7);
            plist->Fixed      = (SBPtr[1] >> 6) & 1;
            plist->Note       = SBPtr[1] & 0x3F;
            plist->FXParam[0] = SBPtr[2];
            plist->FXParam[1] = SBPtr[3];

            SBPtr += 4;
        }
    }

    SongLoaded = true;
    return (true);
}

int8_t ahxplay_Play(int32_t subSongNr)
{
    uint8_t i;
    AHXVoice *v;

    if (!SongLoaded || (subSongNr > Song.SubsongNr))
        return (false);

    PosNr = (subSongNr <= 0) ? 0 : Song.Subsongs[subSongNr - 1];
    PosJump = 0;
    PatternBreak = false;
    MainVolume = 0x40;
    NoteNr = PosJumpNote = 0;
    Tempo = 6;
    StepWaitFrames = 0;
    GetNewPosition = true;
    SongEndReached = false;

#ifdef USE_HIGHPASS
    clearLossyIntegrator(&filterHi);
#endif

#ifdef USE_BLEP
    memset(blep,    0, sizeof (blep_t));
    memset(blepVol, 0, sizeof (blep_t));
#endif

    for (i = 0; i < AMIGA_VOICES; ++i)
    {
        v = &Voices[i];

        memset(v, 0, sizeof (AHXVoice));
        v->TrackOn = true;
        v->TrackMasterVolume = 0x40;

        PaulaSetPeriod(i, 0x88); // dunno about this, but the original asm replayer does it!
        PaulaSetVolume(i, 0);
        PaulaSetData(i, v->VoiceBuffer);
        PaulaSetLength(i, 0x280 / 2);
        PaulaRestartDMA(i);
    }

    samplesPerFrame = (int32_t)(((f_outputFreq / 50.0f) / Song.SpeedMultiplier) + 0.5f);

    Playing = true;
    return (true);
}

static const int16_t midFilterTable[NUM_WAVEFORMS * FILTER_ITERATIONS] =
{
    0xFB77,0xEEC3,0xE407,0xCCDA,0x027B,0x33C7,0x088D,0x1901,0x2351,
    0x3F02,0x3494,0x14F0,0x18CD,0x319B,0x4A69,0x6336,0x7700,0x7F00,
    0x7F00,0x7F03,0x7B89,0x743C,0x6A16,0x5DFC,0x50BB,0x4304,0x3692,
    0x2C6F,0x242F,0x1D77,0x17FE,0x138A,0x0FEA,0x0CF6,0x0A8E,0x0882,
    0x06DB,0x0587,0x0475,0x038D,0x02CC,0x0233,0x01BC,0x014B,0xFD67,
    0xF7DE,0xE7E6,0xDBED,0xCACA,0x3101,0x2591,0x0F6F,0x2099,0x2BEE,
    0x4836,0x1B05,0x0F08,0x21BB,0x4377,0x6533,0x7DA3,0x7F00,0x7EC7,
    0x780E,0x6B20,0x5A61,0x47DD,0x362D,0x28BD,0x1EA3,0x1709,0x1153,
    0x0D07,0x09CB,0x075D,0x056D,0x03FF,0x02D0,0x0212,0x0161,0x0104,
    0x00AD,0x0060,0x0020,0xFFEE,0xFFC9,0xFFB1,0xFFA4,0xFFA1,0xFCBA,
    0xF363,0xE37E,0xCF9E,0xE43D,0x367A,0x1965,0x1752,0x23AD,0x3A63,
    0x41F1,0x17C1,0x0BE8,0x2AA9,0x5553,0x7A8B,0x7F00,0x7D44,0x70C0,
    0x5C86,0x4508,0x2FC9,0x2115,0x16E6,0x0FDA,0x0AF9,0x0798,0x0542,
    0x0384,0x0259,0x0173,0x00DF,0x0089,0x0040,0x0007,0xFFDE,0xFFC6,
    0xFFBB,0xFFBA,0xFFC1,0xFFCC,0xFFD9,0xFFE6,0xFFF2,0xFFFB,0x1378,
    0xEE84,0xE05A,0xC5D4,0x0B4E,0x31B3,0x1313,0x1F4A,0x2616,0x45DF,
    0x2E0E,0x13EB,0x09D8,0x3397,0x672F,0x7F00,0x7EC9,0x7012,0x564D,
    0x3949,0x2460,0x1719,0x0EAA,0x0950,0x05E9,0x038F,0x0224,0x014A,
    0x008F,0x0003,0xFFAA,0xFF7E,0xFF75,0xFF83,0xFF9F,0xFFBF,0xFFDD,
    0xFFF5,0x0006,0x000F,0x0013,0x0013,0x0010,0x000C,0x0008,0x1ADD,
    0xE985,0xDC57,0xC2A3,0x25E9,0x2A8D,0x103D,0x269A,0x2A91,0x4B24,
    0x1FD9,0x10BD,0x0865,0x3C85,0x779A,0x7F00,0x760C,0x599E,0x377B,
    0x2031,0x12AD,0x0AD6,0x0649,0x03A5,0x01F5,0x00DC,0x0051,0x0023,
    0x0002,0xFFEE,0xFFE6,0xFFE7,0xFFEC,0xFFF3,0xFFF9,0xFFFF,0x0002,
    0x0004,0x0004,0x0003,0x0002,0x0001,0x0000,0x0000,0xFFFF,0x097F,
    0xE4D4,0xD636,0xC6FE,0x31B0,0x2314,0x0E82,0x2A8C,0x314E,0x4C62,
    0x1B03,0x0EA1,0x0750,0x4573,0x7F00,0x7F6E,0x66AE,0x3FAE,0x219D,
    0x11BE,0x095D,0x04F1,0x0257,0x011B,0x002D,0xFFA4,0xFF73,0xFF7D,
    0xFFA3,0xFFCF,0xFFF2,0x0008,0x0012,0x0012,0x000E,0x0008,0x0003,
    0x0000,0xFFFE,0xFFFD,0xFFFE,0xFFFE,0xFFFF,0x0000,0x0000,0xF1BA,
    0xE0B8,0xCE39,0xD4B0,0x3539,0x1CAE,0x0D02,0x2C42,0x3A0B,0x4951,
    0x1954,0x0CF7,0x067C,0x4E61,0x7F00,0x77EB,0x5274,0x2978,0x13D3,
    0x0979,0x0487,0x01DD,0x00C4,0x0001,0xFFA3,0xFF93,0xFFAE,0xFFD4,
    0xFFF4,0x0007,0x000E,0x000D,0x0009,0x0004,0x0000,0xFFFE,0xFFFE,
    0xFFFF,0xFFFF,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0xE1AB,
    0xDD5D,0xC593,0xE91A,0x34EE,0x17FB,0x0BAC,0x2C14,0x429E,0x40DA,
    0x1781,0x0BA3,0x05D1,0x574F,0x7F00,0x6DB3,0x3CD8,0x1A34,0x0B48,
    0x04DB,0x0217,0x00BC,0x0020,0xFFD1,0xFFC0,0xFFD1,0xFFEA,0xFFFD,
    0x0007,0x0008,0x0005,0x0003,0x0000,0xFFFF,0xFFFF,0xFFFF,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0xDC89,
    0xDAC4,0xBDC0,0xFEB1,0x32C9,0x14D5,0x0A90,0x2BB8,0x4936,0x3581,
    0x1551,0x0A8F,0x0547,0x603D,0x7F00,0x5FEC,0x2A63,0x1059,0x064E,
    0x026E,0x00B8,0x000F,0xFFC7,0xFFC5,0xFFDE,0xFFF7,0x0005,0x0008,
    0x0006,0x0002,0x0000,0xFFFF,0xFFFF,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0xDE80,
    0xD8C5,0xB789,0x1114,0x2F9E,0x12C9,0x09A8,0x2BDE,0x4D5B,0x2BA2,
    0x1359,0x09A9,0x04D4,0x692B,0x7F00,0x5057,0x1D06,0x09F6,0x036B,
    0x00D4,0x0033,0xFFE2,0xFFD5,0xFFE7,0xFFFA,0x0003,0x0005,0x0003,
    0x0001,0x0000,0xFFFF,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0xE4D0,
    0xD70B,0xB2E4,0x1EB8,0x2BD7,0x1161,0x08E7,0x2D67,0x4F9C,0x2510,
    0x11C9,0x08E8,0x0474,0x7219,0x7C55,0x3F6B,0x133C,0x05D5,0x01C4,
    0x0056,0xFFF9,0xFFE0,0xFFEC,0xFFFB,0x0002,0x0003,0x0002,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0xED7D,
    0xD522,0xB289,0x2800,0x27CE,0x1049,0x0842,0x30EC,0x50A9,0x2153,
    0x1082,0x0842,0x0421,0x7B07,0x73E8,0x2E8C,0x0C60,0x0349,0x0079,
    0x0011,0xFFEA,0xFFEE,0xFFFB,0x0002,0x0002,0x0001,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0xF6E3,
    0xD287,0xB4A8,0x2DF5,0x23D7,0x0F53,0x07B3,0x3641,0x50A6,0x1F47,
    0x0F66,0x07B3,0x03D9,0x7F00,0x6B22,0x20FE,0x079D,0x01C1,0x002D,
    0xFFF5,0xFFF0,0xFFFB,0x0001,0x0001,0x0001,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x001D,
    0xCEB8,0xB9CD,0x3192,0x2037,0x0E6D,0x0736,0x3D2E,0x4F2F,0x1DEA,
    0x0E6C,0x0736,0x039B,0x7F00,0x622C,0x188C,0x04DD,0x00F6,0xFFB9,
    0xFFB2,0xFFEF,0x0008,0x0007,0x0001,0xFFFF,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x08B8,
    0xC94F,0xC47E,0x338E,0x1D17,0x0D96,0x06C8,0x4401,0x4BD3,0x1CA4,
    0x0D90,0x06C8,0x0364,0x7F00,0x5811,0x1100,0x02DB,0x0012,0xFF8B,
    0xFFD8,0x0008,0x0009,0x0002,0xFFFF,0xFFFF,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x1065,
    0xC224,0xD328,0x3460,0x1A87,0x0CD1,0x0667,0x4B18,0x469A,0x1B42,
    0x0CCC,0x0667,0x0333,0x7F00,0x4CC9,0x0A92,0x017C,0xFF6C,0xFFAA,
    0x0002,0x000D,0x0003,0xFFFE,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x16E7,
    0xBA2A,0xDB0C,0x344D,0x187E,0x0C20,0x060F,0x5204,0x402F,0x19D8,
    0x0C1E,0x060F,0x0308,0x7F00,0x40F9,0x0781,0x00DD,0xFFA1,0xFFD9,
    0x0005,0x0005,0x0000,0xFFFF,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x1C0C,
    0xB0D2,0xE7CE,0x337D,0x16EA,0x0B82,0x05C1,0x5814,0x399A,0x1881,
    0x0B82,0x05C1,0x02E1,0x7F00,0x3535,0x04EF,0x0074,0xFFCB,0xFFF1,
    0x0004,0x0002,0xFFFF,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x1FB5,
    0xAD77,0xF515,0x3209,0x15AE,0x0AF4,0x057A,0x5CA5,0x340E,0x174A,
    0x0AF4,0x057A,0x02BD,0x7F00,0x29BF,0x0308,0xFFC8,0xFFC8,0x0004,
    0x0004,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x2170,
    0xB20F,0x01FC,0x300B,0x14AF,0x0A73,0x0539,0x6215,0x2FE7,0x1634,
    0x0A73,0x053A,0x029D,0x7F00,0x1EE1,0x01B1,0xFFDC,0xFFEA,0x0003,
    0x0001,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x2128,
    0xB6D5,0x0E58,0x2D9F,0x13D8,0x09FE,0x04FF,0x68E7,0x2CD8,0x153A,
    0x09FD,0x04FF,0x027F,0x7F00,0x14ED,0x00D4,0xFFA1,0x0000,0x0004,
    0xFFFF,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x1F29,
    0xA209,0x1989,0x2AEA,0x1313,0x0992,0x04C9,0x6FC3,0x2AA6,0x1456,
    0x0992,0x04C9,0x0264,0x7F00,0x0C3B,0x0053,0xFFDD,0x0002,0x0001,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x1BB0,
    0x8888,0x235E,0x2819,0x1258,0x092F,0x0498,0x7023,0x28FE,0x1384,
    0x092F,0x0497,0x024C,0x7F00,0x0780,0xFF65,0xFFF3,0x0004,0xFFFF,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x1740,
    0x808D,0x2BF1,0x255C,0x11A7,0x08D4,0x046A,0x7006,0x2781,0x12C3,
    0x08D4,0x046A,0x0235,0x7F00,0x0423,0xFFB7,0xFFFF,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x1215,
    0x800F,0x338F,0x22E6,0x10FF,0x087F,0x0440,0x6F1E,0x262F,0x120F,
    0x087F,0x043F,0x0220,0x7F00,0x01B2,0xFFEA,0x0001,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0C3C,
    0x8000,0x3B79,0x20EE,0x1062,0x0831,0x0419,0x780C,0x24DF,0x1168,
    0x0831,0x0419,0x020C,0x7F00,0x004B,0xFFFA,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0541,
    0x8000,0x417D,0x1FAB,0x0FD0,0x07E9,0x03F4,0x7F00,0x2398,0x10CE,
    0x07E9,0x03F4,0x01FA,0x7E81,0x0188,0x0005,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0xFD3E,
    0x8000,0x45D7,0x1F45,0x0F49,0x07A4,0x03D2,0x7F00,0x228B,0x103D,
    0x07A5,0x03D2,0x01E9,0x79D0,0x0687,0x007A,0x000A,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0xF416,
    0x8000,0x49EB,0x1FE3,0x0ED7,0x0765,0x03B2,0x7F00,0x21A5,0x0FB6,
    0x0765,0x03B1,0x01D9,0x74CF,0x0C02,0x013C,0x0034,0x000B,0x0003,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0xE94C,
    0x8000,0x4D8B,0x21B2,0x0E9B,0x0729,0x0395,0x7F00,0x1F2E,0x0F31,
    0x0729,0x0394,0x01CB,0x6F7D,0x11D7,0x02DB,0x00CE,0x0042,0x0017,
    0x0008,0x0001,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
    0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0xDBED,
    0x8000,0x506B,0x24C0,0x0F01,0x06F8,0x0379,0x7F00,0x1956,0x0E68,
    0x06F0,0x0379,0x01BF,0x69DB,0x17E0,0x0563,0x0139,0x0087,0x0041,
    0x0021,0x0011,0x0007,0x0004,0x0002,0x0002,0x0002,0x0002,0x0002,
    0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,
    0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0xCE57
};

static const int16_t lowFilterTable[NUM_WAVEFORMS * FILTER_ITERATIONS] =
{
    0x04A4,0x0526,0xFB66,0xEF30,0x9930,0xC5AD,0xF94C,0xFA32,0x09BE,
    0x0E1B,0x5703,0x6B3A,0x83A1,0x8C1C,0x996E,0xAB98,0xC1F8,0xDAA8,
    0xF375,0x0C42,0x2499,0x3BB1,0x50ED,0x63E3,0x744E,0x7F00,0x7F00,
    0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,
    0x7F00,0x7F00,0x7FEC,0x7F58,0x7FEE,0x7F36,0x7F93,0x7FDA,0x22B6,
    0x06E2,0x01FA,0xF97F,0xD0B0,0x9FBF,0xDEBF,0xF750,0x00F7,0x0CD8,
    0x26C6,0x64BB,0x70ED,0x86B8,0x9666,0xAF0A,0xCF0C,0xF0C8,0x127D,
    0x32F3,0x5058,0x6982,0x7DD7,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,
    0x7F00,0x7F00,0x7FFA,0x7FB8,0x7F87,0x7F00,0x7F99,0x7F00,0x7F4B,
    0x7F7E,0x7F9D,0x7FA9,0x7FA7,0x7F9B,0x7F88,0x7F70,0x7F56,0x1827,
    0x085D,0xFD8A,0xF58E,0xAAA0,0xB87E,0xE9B1,0xF78F,0x089F,0x1097,
    0x44D0,0x676B,0x7417,0x8ABF,0xA3D4,0xCA53,0xF4FD,0x1F5D,0x4681,
    0x6766,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7FE2,
    0x7F97,0x7F64,0x7FF4,0x7F25,0x7F5A,0x7F75,0x7F7C,0x7F74,0x7F62,
    0x7F4C,0x7F34,0x7F1E,0x7F0C,0x7EFE,0x7EF4,0x7EEE,0x7EEC,0x0E14,
    0x08C8,0xFA29,0xEA14,0x9750,0xCB17,0xED77,0xFA92,0x0D73,0x1B3D,
    0x5BC7,0x6C4C,0x7626,0x8FB7,0xB465,0xE741,0x1ACD,0x4A39,0x7012,
    0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F9B,
    0x7FE7,0x7FF5,0x7FDA,0x7FA9,0x7F70,0x7F3C,0x7F11,0x7EF4,0x7EE2,
    0x7EDC,0x7EDD,0x7EE2,0x7EE9,0x7EF1,0x7EF8,0x7EFD,0x7F01,0x1F29,
    0x07B7,0xF8BB,0xD9E8,0x9ADB,0xD85A,0xEF6F,0x0056,0x105E,0x2993,
    0x6293,0x6F41,0x779A,0x95A1,0xC7C2,0x0448,0x3E6E,0x6CE0,0x7F00,
    0x7F00,0x7F00,0x7F00,0x7F00,0x7FDE,0x7F77,0x7FFF,0x7F13,0x7F29,
    0x7F2D,0x7F27,0x7F1B,0x7F0F,0x7F04,0x7EFD,0x7EF9,0x7EF8,0x7EF9,
    0x7EFA,0x7EFC,0x7EFE,0x7EFF,0x7F00,0x7F00,0x7F00,0x7EFF,0x38C1,
    0x056D,0xF814,0xC623,0xA713,0xE15B,0xF142,0x07C2,0x1262,0x38E2,
    0x63C5,0x715B,0x78AE,0x9C7C,0xDBB2,0x2144,0x5DAB,0x7F00,0x7F00,
    0x7F00,0x7F00,0x7F00,0x7F00,0x7FC4,0x7FFA,0x7FD7,0x7F8E,0x7F43,
    0x7F0A,0x7EEA,0x7EDE,0x7EE0,0x7EE8,0x7EF2,0x7EFB,0x7F00,0x7F03,
    0x7F03,0x7F02,0x7F01,0x7F00,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x44FF,
    0x022D,0xF638,0xB1EC,0xB3D3,0xE6DC,0xF2F4,0x1027,0x1555,0x4964,
    0x65A0,0x7308,0x7983,0xA447,0xEFC0,0x3BF3,0x755D,0x7F00,0x7F00,
    0x7F00,0x7F00,0x7F00,0x7FA2,0x7FBC,0x7F8E,0x7F4A,0x7F13,0x7EF2,
    0x7EE6,0x7EE8,0x7EF0,0x7EF8,0x7EFE,0x7F01,0x7F02,0x7F01,0x7F00,
    0x7EFF,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x3F9E,
    0xFE6E,0xF23E,0xA271,0xBEFF,0xEA01,0xF459,0x186B,0x1AB5,0x58FD,
    0x6858,0x745C,0x7A2D,0xAD04,0x0454,0x552E,0x7F00,0x7F00,0x7F00,
    0x7F00,0x7FB7,0x7F40,0x7F6E,0x7F58,0x7F2D,0x7F08,0x7EF5,0x7EF0,
    0x7EF4,0x7EF9,0x7EFD,0x7F00,0x7F00,0x7F00,0x7EFF,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x347C,
    0xFAB9,0xED47,0x9AA4,0xC870,0xEBE1,0xF572,0x1E07,0x2265,0x6469,
    0x6AD4,0x7570,0x7AB7,0xB6B2,0x16F0,0x681C,0x7F00,0x7F00,0x7F00,
    0x7FEB,0x7F45,0x7F6C,0x7F4A,0x7F1C,0x7EFC,0x7EF1,0x7EF3,0x7EF9,
    0x7EFE,0x7F00,0x7F00,0x7F00,0x7EFF,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x28BB,
    0xF78E,0xE420,0x99DD,0xD05D,0xED36,0xF656,0x2213,0x2B5A,0x6A73,
    0x6CB8,0x7655,0x7B2A,0xC151,0x2A7D,0x790E,0x7F00,0x7F00,0x7F00,
    0x7F00,0x7F48,0x7F3B,0x7F18,0x7EFF,0x7EF6,0x7EF7,0x7EFB,0x7EFE,
    0x7F00,0x7F00,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x1E75,
    0xF544,0xD834,0x9DAA,0xD6E9,0xEE5D,0xF717,0x24B5,0x34A1,0x6D21,
    0x6E35,0x7717,0x7B8B,0xCCE1,0x3DC8,0x7F00,0x7F00,0x7F00,0x7FCB,
    0x7F26,0x7F30,0x7F16,0x7F00,0x7EF9,0x7EFA,0x7EFC,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x16D0,
    0xF3F4,0xCBF1,0xA3BF,0xDC27,0xEF79,0xF7BC,0x26A4,0x3DE3,0x6E84,
    0x6F7A,0x77BC,0x7BDE,0xD962,0x4F0B,0x7F00,0x7F00,0x7F00,0x7F00,
    0x7F26,0x7F16,0x7F02,0x7EFB,0x7EFB,0x7EFE,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x11F0,
    0xF371,0xC1E6,0xAAA0,0xE033,0xF08A,0xF84C,0x28A1,0x4735,0x6FC4,
    0x7098,0x784C,0x7C25,0xE441,0x5DE5,0x7F00,0x7F00,0x7FE9,0x7F17,
    0x7F16,0x7F05,0x7EFC,0x7EFC,0x7EFD,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x0F7C,
    0xF342,0xB6DF,0xB189,0xE338,0xF187,0xF8C9,0x2B06,0x509A,0x7144,
    0x7193,0x78C8,0x7C63,0xED84,0x6A6B,0x7F00,0x7F00,0x7F88,0x7F73,
    0x7F12,0x7EEF,0x7EF4,0x7EFE,0x7F00,0x7F00,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x0F12,
    0xF2B4,0xAA33,0xB822,0xE56F,0xF269,0xF936,0x2EE9,0x59B0,0x72F5,
    0x726E,0x7936,0x7C9A,0xF717,0x75F2,0x7F00,0x7F00,0x7FBB,0x7F39,
    0x7EF0,0x7EF0,0x7EFD,0x7F01,0x7F00,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x1051,
    0xF0E3,0xA18C,0xBE4C,0xE714,0xF330,0xF998,0x33B3,0x61CB,0x74A7,
    0x7332,0x7998,0x7CCB,0x00FB,0x7FF6,0x7F00,0x7FED,0x7F8C,0x7EFC,
    0x7EEA,0x7EFB,0x7F01,0x7F00,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x12E5,
    0xECCF,0x98BC,0xC3FC,0xE858,0xF3E1,0xF9EF,0x39A7,0x6850,0x7630,
    0x73E0,0x79EF,0x7CF7,0x0B2F,0x7F00,0x7F00,0x7F91,0x7F44,0x7EF7,
    0x7EF5,0x7EFE,0x7F00,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x167B,
    0xE76E,0x93CA,0xC92C,0xE964,0xF47D,0xFA3E,0x40A5,0x6D08,0x778A,
    0x747D,0x7A3E,0x7D1E,0x15B3,0x7F00,0x7F00,0x7F50,0x7F1D,0x7EF9,
    0x7EFB,0x7EFF,0x7EFF,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x1AF2,
    0xDD31,0x9374,0xCDD7,0xEA52,0xF50B,0xFA85,0x484D,0x700F,0x78BA,
    0x750B,0x7A84,0x7D42,0x2088,0x7F00,0x7F00,0x7F6F,0x7EFF,0x7EF6,
    0x7EFE,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x1FAB,
    0xD337,0x963B,0xD1F3,0xEB33,0xF58C,0xFAC5,0x4D79,0x721D,0x79CB,
    0x758C,0x7AC5,0x7D62,0x2BAD,0x7F00,0x7F00,0x7F31,0x7EFC,0x7EFC,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x241F,
    0xCC6D,0x9B30,0xD57A,0xEC0B,0xF601,0xFB00,0x5244,0x73C4,0x7AC4,
    0x7601,0x7B00,0x7D7F,0x3722,0x7F00,0x7FA9,0x7F11,0x7EF6,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x280C,
    0xBE21,0xA145,0xD867,0xECDA,0xF66D,0xFB36,0x55B1,0x7546,0x7BA8,
    0x766D,0x7B36,0x7D9A,0x42E8,0x7F00,0x7F45,0x7F02,0x7EFC,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x2B18,
    0xA865,0xA7BA,0xDAC1,0xED9F,0xF6CF,0xFB67,0x5207,0x76C6,0x7C7B,
    0x76CF,0x7B67,0x7DB3,0x4EFE,0x7F00,0x7F4F,0x7EF5,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x2D13,
    0xB2BE,0xAE1E,0xDC9C,0xEE56,0xF72B,0xFB95,0x4C1E,0x7845,0x7D3D,
    0x772B,0x7B95,0x7DC9,0x5B64,0x7F00,0x7F17,0x7EFC,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x2E1B,
    0xCE29,0xB467,0xDE0E,0xEEFF,0xF77F,0xFBBF,0x4843,0x79BD,0x7DF0,
    0x777F,0x7BBE,0x7DDE,0x681B,0x7F00,0x7F02,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x2E32,
    0xE25E,0xBB9A,0xDF40,0xEF9C,0xF7CD,0xFBE6,0x68ED,0x7B21,0x7E96,
    0x77CD,0x7BE6,0x7DF1,0x7522,0x7F49,0x7EFE,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,
    0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x7EFF,0x2D4F,
    0xF4E0,0xC241,0xE05F,0xF02D,0xF817,0xFC0A,0x77E5,0x7C6A,0x7F31,
    0x7817,0x7C0B,0x7E05,0x7F00,0x7F0C,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x2B1D,
    0xEB4A,0xC91D,0xE1A6,0xF0B7,0xF85A,0xFC2C,0x6D81,0x7DA5,0x7FC1,
    0x785A,0x7C2D,0x7E16,0x7F00,0x7FD9,0x7F18,0x7F01,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x2782,
    0xE163,0xCFFB,0xE34C,0xF13A,0xF89A,0xFC4C,0x615E,0x7ED9,0x7F00,
    0x7899,0x7C4B,0x7E26,0x7F00,0x7F00,0x7F44,0x7F0F,0x7F03,0x7F00,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x2237,
    0xD728,0xD6A1,0xE592,0xF1CA,0xF8D5,0xFC6B,0x5DFA,0x7F58,0x7F00,
    0x78D6,0x7C6A,0x7E35,0x7F00,0x7F00,0x7FD9,0x7F4F,0x7F19,0x7F08,
    0x7F02,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,
    0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x7EFE,0x1A9B,
    0xCC9C,0xDCAB,0xE8B3,0xF2A3,0xF910,0xFC86,0x7B5E,0x7DF6,0x7F00,
    0x790C,0x7C86,0x7E44,0x7F00,0x7F00,0x7F00,0x7F74,0x7F3D,0x7F1D,
    0x7F0E,0x7F06,0x7F02,0x7F01,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,
    0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,
    0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x7F00,0x12E3
};